/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is JSIRC Library
 *
 * The Initial Developer of the Original Code is New Dimensions Consulting,
 * Inc. Portions created by New Dimensions Consulting, Inc. are
 * Copyright (C) 1999 New Dimenstions Consulting, Inc. All
 * Rights Reserved.
 *
 * Contributor(s):
 *  Robert Ginda, rginda@ndcico.com, original author
 *  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
 *
 * depends on utils.js, and XPCOM/XPConnect in the JS environment
 *
 */

// Changes by David Costanzo
// * Make this not depend on utils.js so it can be used independently.
// * Put everything into a Dict namespace so that changes to jslib won't
//   affect this code (and vice-versa) if it is loaded into the same 
//   JavaScript namespace.
// * Always make HAS_NSPR_EVENTQ=true
"use strict";

if ("undefined" == typeof(DictConnection)) {
    var DictConnection = {};
};

var EXPORTED_SYMBOLS = ['DictConnection'];

DictConnection.JSENV = {
    HAS_DOCUMENT        : typeof document == "object",
    HAS_STREAM_PROVIDER : "nsIStreamProvider" in Components.interfaces,

    // HACK:
    //   This was originally (typeof document == "object"), but I have
    //   change it to true because the dict protocol handler
    //   doesn't work otherwise.  I supsect it should be true, anyway,
    //   and that testing if 'document' is defined is not a good way
    //   to determine if we have an event queue.
    HAS_NSPR_EVENTQ : true
};


DictConnection.toScriptableInputStream = function(i)
{
    var si = Components.classes["@mozilla.org/scriptableinputstream;1"];
    
    si = si.createInstance();
    si = si.QueryInterface(Components.interfaces.nsIScriptableInputStream);
    si.init(i);

    return si;
}

DictConnection.CBSConnection = function()
{
    var sockServiceClass =
        Components.classesByID["{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"];
    
    if (!sockServiceClass)
        throw ("Couldn't get socket service class.");
    
    var sockService = sockServiceClass.getService();
    if (!sockService)
        throw ("Couldn't get socket service.");

    this._sockService = sockService.QueryInterface(
        Components.interfaces.nsISocketTransportService);

    this.wrappedJSObject = this;
}

DictConnection.CBSConnection.prototype = {

    connect : function(host, port, bind, tcp_flag, observer)
    {
        if (typeof tcp_flag == "undefined")
            tcp_flag = false;
    
        this.host = host.toLowerCase();
        this.port = port;
        this.bind = bind;
        this.tcp_flag = tcp_flag;

        if (DictConnection.JSENV.HAS_STREAM_PROVIDER) {
            this._transport = this._sockService.createTransport(host, port, null, 0, 0);

            if (!this._transport)
                throw "Error creating transport.";

            if (DictConnection.JSENV.HAS_NSPR_EVENTQ) {
                // we've got an event queue, so start up an async write.
                this._streamProvider = new DictStreamProvider(observer);
                this._write_req = this._transport.asyncWrite(this._streamProvider, this, 0, -1, 0);
            }
            else {
                /* no nspr event queues in this environment, we can't use async calls,
                 * so set up the streams. */
                this._outputStream = this._transport.openOutputStream(0, -1, 0);
                if (!this._outputStream)
                    throw "Error getting output stream. (no async, stream provider)";

                this._inputStream = this._transport.openInputStream (0, -1, 0);
                if (!this._inputStream)
                    throw "Error getting input stream. (no async, stream provider)";
            }
        }
        else {
            // use new necko interfaces (bug 111479)
            this._transport = this._sockService.createTransport(null, 0, host, port, null);
            if (!this._transport)
                throw "Error creating transport.";

            var openFlags;
            if (DictConnection.JSENV.HAS_NSPR_EVENTQ)
                openFlags = 0;
            else
                openFlags = Components.interfaces.nsITransport.OPEN_BLOCKING;

            this._outputStream = this._transport.openOutputStream(openFlags, 4096, 32);
            if (!this._outputStream)
                throw "Error getting output stream. (no stream provider)";

            this._inputStream = this._transport.openInputStream(openFlags, 0, 0);
            if (!this._inputStream)
                throw "Error getting input stream. (no stream provider)";
        }

        this.connectDate = new Date();
        this.isConnected = true;

        return this.isConnected;
    },

    disconnect : function()
    {
        if ("_inputStream" in this && this._inputStream) {
            this._inputStream.close();
        }
        if ("_outputStream" in this && this._outputStream) {
            this._outputStream.close();
        }

        this.isConnected = false;
    },

    sendData : function(str)
    {
        if (!this.isConnected)
            throw "Not Connected.";

        if (DictConnection.JSENV.HAS_NSPR_EVENTQ && DictConnection.JSENV.HAS_STREAM_PROVIDER)
            this.asyncWrite(str);
        else
            this.sendDataNow(str);
    },
    
    readData : function(timeout, count)
    {
        if (!this.isConnected)
            throw "Not Connected.";

        var rv;

        try {
            rv = this._scriptableInputStream.read(count);
        }
        catch (ex) {
            //dd("*** Caught " + ex + " while reading.")
            this.isConnected = false;
            throw ex;
        }

        return rv;
    },

    startAsyncRead : function(observer)
    {
        if (DictConnection.JSENV.HAS_STREAM_PROVIDER) {
            this._transport.asyncRead(new DictConnection.StreamListener(observer), this, 0, -1, 0);
        }
        else {
            var cls = Components.classes["@mozilla.org/network/input-stream-pump;1"];
            var pump = cls.createInstance(Components.interfaces.nsIInputStreamPump);
            pump.init(this._inputStream, -1, -1, 0, 0, false);
            pump.asyncRead(new DictConnection.StreamListener(observer), this);
        }
    },

    asyncWrite : function(str)
    {
        this._streamProvider.pendingData += str;
        if (this._streamProvider.isBlocked) {
            this._write_req.resume();
            this._streamProvider.isBlocked = false;
        }
    },

    hasPendingWrite : function()
    {
        if (DictConnection.JSENV.HAS_STREAM_PROVIDER)
            return (this._streamProvider.pendingData != "");
        else
            return false; // data already pushed to necko
    },

    sendDataNow : function(str)
    {
        var rv = false;
    
        try {
            this._outputStream.write(str, str.length);
            rv = true;
        }
        catch (ex) {
            //dd ("*** Caught " + ex + " while sending.")
            this.isConnected = false;
            throw ex;
        }
    
        return rv;
    },

    _notimpl : function()
    {
        throw "Not Implemented.";
    }
};

if (!DictConnection.JSENV.HAS_NSPR_EVENTQ) {
    DictConnection.CBSConnection.prototype.startAsyncRead = DictConnection.CBSConnection.prototype._notimpl;
    DictConnection.CBSConnection.prototype.asyncWrite     = DictConnection.CBSConnection.prototype._notimpl;
}
else if (DictConnection.JSENV.HAS_STREAM_PROVIDER) {
    DictConnection.CBSConnection.prototype.sendDataNow = DictConnection.CBSConnection.prototype._notimpl;
}
else {
    DictConnection.CBSConnection.prototype.asyncWrite = DictConnection.CBSConnection.prototype._notimpl;
}

DictConnection.StreamProvider = function(observer)
{
    this._observer = observer;
}

DictConnection.StreamProvider.prototype = {
    pendingData : "",
    isBlocked   : true,

    onDataWritable : function(request, ctxt, ostream, offset, count)
    {
        if ("isClosed" in this && this.isClosed)
            throw Components.results.NS_BASE_STREAM_CLOSED;

        if (!this.pendingData) {
            this.isBlocked = true;

            /* this is here to support pre-XPCDOM builds (0.9.0 era), which
             * don't have this result code mapped. */
            if (!Components.results.NS_BASE_STREAM_WOULD_BLOCK)
                throw 2152136711;
        
            throw Components.results.NS_BASE_STREAM_WOULD_BLOCK;
        }
    
        var len = ostream.write (this.pendingData, this.pendingData.length);
        this.pendingData = this.pendingData.substr (len);
    },

    onStartRequest : function(request, ctxt)
    {
    },

    onStopRequest : function(request, ctxt, status)
    {
        if (this._observer)
            this._observer.onStreamClose(status);
    }
};

DictConnection.StreamListener = function(observer)
{
    this._observer = observer;
}

DictConnection.StreamListener.prototype = {
    onStartRequest : function(request, ctxt)
    {
    },

    onStopRequest : function(request, ctxt, status)
    {
        if (this._observer)
            this._observer.onStreamClose(status);
    },

    onDataAvailable : function(request, ctxt, inStr, sourceOffset, count)
    {
        ctxt = ctxt.wrappedJSObject;
        if (!ctxt) {
            //dd ("*** Can't get wrappedJSObject from ctxt in " +
            //    "DictStreamListener.onDataAvailable ***");
            return;
        }

        if (!("_scriptableInputStream" in ctxt)) {
            ctxt._scriptableInputStream = DictConnection.toScriptableInputStream(inStr);
        }

        if (this._observer) {
            this._observer.onStreamDataAvailable(request, inStr, sourceOffset, count);
        }
    }
};
